@@ -9,6 +9,7 @@ This file defines various text generation classes and methods
 import math
 import random
 
+import logging
 import numpy as np
 import pandas as pd
 
@@ -203,7 +204,7 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
               If the `escapeSpecialChars` option is set to True, then the following char only has its special
               meaning when preceded by an escape.
 
-              Some options must be always escaped for example ``\\0``, ``\\v``, ``\\n`` and ``\\w``.
+              Some options must be always escaped for example  ``\\v``, ``\\n`` and ``\\w``.
 
               A special case exists for ``\\v`` - if immediately followed by a digit 0 - 9, the underlying base value
               is interpreted as an array of values and the nth element is retrieved where `n` is the digit specified.
@@ -219,15 +220,13 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
     when applied to the values zero to 999. This conforms to the preferred style going forward
 
     """
-
     def __init__(self, template, escapeSpecialChars=False, extendedWordList=None):
         assert template is not None, "`template` must be specified"
         super().__init__()
 
         self._template = template
         self._escapeSpecialMeaning = bool(escapeSpecialChars)
-        template_str0 = self._template
-        self._templates = [x.replace('$__sep__', '|') for x in template_str0.replace(r'\|', '$__sep__').split('|')]
+        self._templates = self._splitTemplates(self._template)
         self._wordList = np.array(extendedWordList if extendedWordList is not None else _WORDS_LOWER)
         self._upperWordList = np.array([x.upper() for x in extendedWordList]
                                        if extendedWordList is not None else _WORDS_UPPER)
@@ -243,18 +242,80 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
         self._np_letters_all = np.array(_LETTERS_ALL)
         self._lenWords = len(self._wordList)
 
-        # get the template metadata
-        template_info =  [self._prepareTemplateStrings(template, escapeSpecialMeaning=escapeSpecialChars)
+        # mappings must be mapping from string to tuple(length of mappings, mapping array or list)
+        self._templateMappings = {
+            'a': (26, self._np_letters_lower),
+            'A': (26, self._np_letters_upper),
+            'x': (16, self._np_hex_lower),
+            'X': (16, self._np_hex_upper),
+            'd': (10, self._np_digits_zero),
+            'D': (9, self._np_digits_non_zero),
+            'k': (36, self._np_alnum_lower),
+            'K': (36, self._np_alnum_upper)
+        }
+
+        # ensure that each mapping is mapping from string to list or numpy array
+        for k, v in self._templateMappings.items():
+            assert (k is not None) and isinstance(k, str) and len(k) > 0, "key must be non-empty string"
+            assert v is not None and isinstance(v, tuple) and len(v) == 2, "value must be tuple of length 2"
+            mapping_length, mappings = v
+            assert isinstance(mapping_length, int), "mapping length must be of type int"
+            assert isinstance(mappings, (list, np.ndarray)),\
+                "mappings are lists or numpy arrays"
+            assert mapping_length == 0 or len(mappings) == mapping_length, "mappings must match mapping_length"
+
+        self._templateEscapedMappings = {
+            'n': (256, None),
+            'N': (65536, None),
+            'w': (self._lenWords, self._wordList),
+            'W': (self._lenWords, self._upperWordList)
+        }
+
+        # ensure that each escaped mapping is mapping from string to None, list or numpy array
+        for k, v in self._templateEscapedMappings.items():
+            assert (k is not None) and isinstance(k, str) and len(k) > 0, "key must be non-empty string"
+            assert v is not None and isinstance(v, tuple) and len(v) == 2, "value must be tuple of length 2"
+            mapping_length, mappings = v
+            assert isinstance(mapping_length, int), "mapping length must be of type int"
+            assert mappings is None or isinstance(mappings, (list, np.ndarray)),\
+                "mappings are lists or numpy arrays"
+
+            # for escaped mappings, the mapping can be None in which case the mapping is to the number itself
+            # i.e mapping[4] = 4
+            assert mappings is None or len(mappings) == mapping_length, "mappings must match mapping_length"
+
+        # get the template metadata - this will be list of metadata entries for each template
+        # for each template, metadata will be tuple of number of placeholders followed by list of random bounds
+        # to be computed when replacing non static placeholder
+        template_info = [self._prepareTemplateStrings(template, escapeSpecialMeaning=escapeSpecialChars)
                                     for template in self._templates]
-        self._max_placeholders = max([ x[0] for x in template_info])
-        self._max_rnds_needed = max([ len(x[1]) for x in template_info])
+
+        logger = logging.getLogger(__name__)
+
+        #if logger.isEnabledFor(logging.DEBUG):
+        for ix, ti in template_info:
+            logger.info(f"templates - {ix} {ti}")
+
+        self._max_placeholders = max([ x[0] for x in template_info])  # pylint: disable=consider-using-generator
+        self._max_rnds_needed = max([ len(x[1]) for x in template_info])  # pylint: disable=consider-using-generator
         self._placeholders_needed = [ x[0] for x in template_info]
         self._template_rnd_bounds = [ x[1] for x in template_info]
 
-
     def __repr__(self):
         return f"TemplateGenerator(template='{self._template}')"
 
+    def _splitTemplates(self, templateStr):
+        """ Split template string into individual template strings
+
+        :param templateStr: template string
+        :return: list of individual template strings
+
+
+        """
+        tmp_template = templateStr.replace(r'\\', '$__escape__').replace(r'\|', '$__sep__')
+        results = [x.replace('$__escape__', r'\\').replace('$__sep__', '|') for x in tmp_template.split('|')]
+        return results
+
     @property
     def templates(self):
         """ Get effective templates for text generator"""
@@ -319,57 +380,17 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
                 # retval.append(str(baseValue[val_index]))
                 num_placeholders += 1
                 use_value = False
-            elif char == 'x' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(16)
+            elif (char in self._templateMappings) and (not escape) ^ escapeSpecialMeaning:
+                # handle case for ['a','A','k', 'K', 'x', 'X']
+                bound, mappingArr = self._templateMappings[char]
+                retval.append(bound)
                 num_placeholders += 1
-                # used for retval.append(_HEX_LOWER[self._getRandomInt(0, 15, rndGenerator)])
-            elif char == 'X' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(16)
-                num_placeholders += 1
-                # retval.append(_HEX_UPPER[self._getRandomInt(0, 15, rndGenerator)])
-            elif char == 'd' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(10)
-                num_placeholders += 1
-                # retval.append(_DIGITS_ZERO[self._getRandomInt(0, 9, rndGenerator)])
-            elif char == 'D' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(9)
-                num_placeholders += 1
-                # retval.append(_DIGITS_NON_ZERO[self._getRandomInt(0, 8, rndGenerator)])
-            elif char == 'a' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(26)
-                num_placeholders += 1
-                # retval.append(_LETTERS_LOWER[self._getRandomInt(0, 25, rndGenerator)])
-            elif char == 'A' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(26)
-                num_placeholders += 1
-                # retval.append(_LETTERS_UPPER[self._getRandomInt(0, 25, rndGenerator)])
-            elif char == 'k' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(26)
-                num_placeholders += 1
-                # retval.append(_ALNUM_LOWER[self._getRandomInt(0, 35, rndGenerator)])
-            elif char == 'K' and (not escape) ^ escapeSpecialMeaning:
-                retval.append(36)
-                num_placeholders += 1
-                # retval.append(_ALNUM_UPPER[self._getRandomInt(0, 35, rndGenerator)])
-            elif char == 'n' and escape:
-                retval.append(256)
-                num_placeholders += 1
-                # retval.append(str(self._getRandomInt(0, 255, rndGenerator)))
                 escape = False
-            elif char == 'N' and escape:
-                retval.append(65536)
+            elif (char in self._templateEscapedMappings) and escape:
+                # handle case for ['n', 'N', 'w', 'W']
+                bound, mappingArr = self._templateEscapedMappings[char]
+                retval.append(bound)
                 num_placeholders += 1
-                # retval.append(str(self._getRandomInt(0, 65535, rndGenerator)))
-                escape = False
-            elif char == 'W' and escape:
-                retval.append(self._lenWords)
-                num_placeholders += 1
-                # retval.append(self._upperWordList[self._getRandomWordOffset(self._lenWords, rndGenerator=rndGenerator)])
-                escape = False
-            elif char == 'w' and escape:
-                retval.append(self._lenWords)
-                num_placeholders += 1
-                # retval.append(self._wordList[self._getRandomWordOffset(self._lenWords, rndGenerator=rndGenerator)])
                 escape = False
             elif char == 'v' and escape:
                 escape = False
@@ -410,12 +431,23 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
         `_escapeSpecialMeaning` parameter allows for backwards compatibility with old style syntax while allowing
         for preferred new style template syntax. Specify as True to force escapes for special meanings,.
 
+        .. note::
+                Both `placeholders` and `rnds` are numpy masked arrays. If there are multiple templates in the template
+                generation source template, then this method will be called multiple times with each of
+                the distinct templates passed and the `placeholders` and `rnds` arrays masked so that the each call
+                will apply the template to rows to which that template applies.
+
+                The template may be the empty string.
+
         """
         assert baseValue.shape[0] == placeholders.shape[0]
         assert baseValue.shape[0] == rnds.shape[0]
 
         _cached_values = {}
 
+        regularKeys = self._templateMappings.keys()
+        escapedKeys = self._templateEscapedMappings.keys()
+
         def _get_values_as_np_array():
             """Get baseValue which is pd.Series or Dataframe as a numpy array and cache it"""
             if "np_values" not in _cached_values:
@@ -428,11 +460,11 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
             cache_key = f"v_{elem}"
             if cache_key not in _cached_values:
                 np_values = _get_values_as_np_array()
-                #element_values = []
-                element_values = np.ndarray( np_values.shape[0], dtype=np_values.dtype)
+                # element_values = []
+                element_values = np.ndarray(np_values.shape[0], dtype=np_values.dtype)
 
                 for x in range(baseValue.shape[0]):
-                    #element_values.append(baseValue[x][elem])
+                    # element_values.append(baseValue[x][elem])
                     element_values[x] = baseValue[x][elem]
                 _cached_values[cache_key] = element_values
 
@@ -444,6 +476,18 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
         num_placeholders = 0
         rnd_offset = 0
 
+        unmasked_rows = None  # unmasked_rows is None, indicates that all rows are unmasked
+
+        assert isinstance(placeholders, np.ma.MaskedArray), "expecting MaskArray"
+
+        # if template is empty, then nothing needs to be done
+        if template_len > 0 and isinstance(placeholders, np.ma.MaskedArray):
+            active_rows = ~placeholders.mask
+            unmasked_rows = active_rows[:, 0]
+
+            if np.all(active_rows):
+                unmasked_rows = None
+
         # in the following code, the construct `(not escape) ^ self._escapeSpecialMeaning` means apply
         # special meaning if either escape is not true or the option `self._escapeSpecialMeaning` is true.
         # This corresponds to the logical xor operation
@@ -456,75 +500,40 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
             elif use_value and ('0' <= char <= '9'):
                 val_index = int(char)
                 placeholders[:, num_placeholders] = _get_values_subelement(val_index)
-                #placeholders[:, num_placeholders] = pd_base_values.apply(lambda x: str(x[val_index]))
+                # placeholders[:, num_placeholders] = pd_base_values.apply(lambda x: str(x[val_index]))
                 num_placeholders += 1
                 use_value = False
-            elif char == 'x' and (not escape) ^ escapeSpecialMeaning:
+            elif char in regularKeys and (not escape) ^ escapeSpecialMeaning:
                 # note vectorized lookup - `rnds[:, rnd_offset]` will get vertical column of
                 # random numbers from `rnds` 2d array
-                placeholders[:, num_placeholders] = self._np_hex_lower[rnds[:, rnd_offset]]
+                bound, valueMappings = self._templateMappings[char]
+
+                if unmasked_rows is not None:
+                    placeholders[unmasked_rows, num_placeholders] = valueMappings[rnds[unmasked_rows, rnd_offset]]
+                else:
+                    placeholders[:, num_placeholders] = valueMappings[rnds[:, rnd_offset]]
+
                 num_placeholders += 1
                 rnd_offset = rnd_offset + 1
+                escape = False
                 # used for retval.append(_HEX_LOWER[self._getRandomInt(0, 15, rndGenerator)])
-            elif char == 'X' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_hex_upper[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_HEX_UPPER[self._getRandomInt(0, 15, rndGenerator)])
-            elif char == 'd' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_digits_zero[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_DIGITS_ZERO[self._getRandomInt(0, 9, rndGenerator)])
-            elif char == 'D' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_digits_non_zero[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_DIGITS_NON_ZERO[self._getRandomInt(0, 8, rndGenerator)])
-            elif char == 'a' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_letters_lower[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_LETTERS_LOWER[self._getRandomInt(0, 25, rndGenerator)])
-            elif char == 'A' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_letters_upper[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_LETTERS_UPPER[self._getRandomInt(0, 25, rndGenerator)])
-            elif char == 'k' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_alnum_lower[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_ALNUM_LOWER[self._getRandomInt(0, 35, rndGenerator)])
-            elif char == 'K' and (not escape) ^ escapeSpecialMeaning:
-                placeholders[:, num_placeholders] = self._np_alnum_upper[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(_ALNUM_UPPER[self._getRandomInt(0, 35, rndGenerator)])
-            elif char == 'n' and escape:
-                placeholders[:, num_placeholders] = rnds[:, rnd_offset]
+            elif char in escapedKeys and escape:
+                bound, valueMappings = self._templateEscapedMappings[char]
+
+                if valueMappings is not None:
+                    if unmasked_rows is not None:
+                        placeholders[unmasked_rows, num_placeholders] = valueMappings[rnds[unmasked_rows, rnd_offset]]
+                    else:
+                        placeholders[:, num_placeholders] = valueMappings[rnds[:, rnd_offset]]
+                else:
+                    if unmasked_rows is not None:
+                        placeholders[unmasked_rows, num_placeholders] = rnds[unmasked_rows, rnd_offset]
+                    else:
+                        placeholders[:, num_placeholders] = rnds[:, rnd_offset]
                 num_placeholders += 1
                 rnd_offset = rnd_offset + 1
                 # retval.append(str(self._getRandomInt(0, 255, rndGenerator)))
                 escape = False
-            elif char == 'N' and escape:
-                placeholders[:, num_placeholders] = rnds[:, rnd_offset]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(str(self._getRandomInt(0, 65535, rndGenerator)))
-                escape = False
-            elif char == 'W' and escape:
-                placeholders[:, num_placeholders] = self._upperWordList[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(self._upperWordList[self._getRandomWordOffset(self._lenWords, rndGenerator=rndGenerator)])
-                escape = False
-            elif char == 'w' and escape:
-                placeholders[:, num_placeholders] = self._wordList[rnds[:, rnd_offset]]
-                num_placeholders += 1
-                rnd_offset = rnd_offset + 1
-                # retval.append(self._wordList[self._getRandomWordOffset(self._lenWords, rndGenerator=rndGenerator)])
-                escape = False
             elif char == 'v' and escape:
                 escape = False
                 if following_char is not None and ('0' <= following_char <= '9'):
@@ -628,24 +637,24 @@ class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
         masked_matrices = [masked_placeholders, masked_rnds]
 
         # test logic for template expansion
-        for x in range(len(self._templates)):
+        for x in range(len(self._templates)):  # pylint: disable=consider-using-enumerate
             masked_placeholders[template_choices_t != x, :] = np.ma.masked
             masked_rnds[template_choices_t != x, :] = np.ma.masked
-            #masked_base_values[template_choices_t != x] = np.ma.masked
+            # masked_base_values[template_choices_t != x] = np.ma.masked
 
             # harden mask, preventing modifications
             for m in masked_matrices:
                 np.ma.harden_mask(m)
 
-            # expand values into placeholders
+            # expand values into placeholders without affect masked values
             #self._applyTemplateStringsForTemplate(v.to_numpy(dtype=np.object_), #masked_base_values,
             self._applyTemplateStringsForTemplate(v,
-                                                    #masked_base_values,
-                                                    self._templates[x],
-                                                    masked_placeholders,
-                                                    masked_rnds,
-                                                    escapeSpecialMeaning=self._escapeSpecialMeaning
-                                                    )
+                                                  # masked_base_values,
+                                                  self._templates[x],
+                                                  masked_placeholders,
+                                                  masked_rnds,
+                                                  escapeSpecialMeaning=self._escapeSpecialMeaning
+                                                  )
 
             # soften and clear mask, allowing modifications
             for m in masked_matrices:
@@ -733,7 +742,7 @@ class ILText(TextGenerator):  # lgtm [py/missing-equals]
 
         # set up python list of all words so that we dont have to convert between numpy and python representations
         self._allWordsSize = all_words.size
-        self._wordsAsPythonStrings = list([str(x) for x in all_words])
+        self._wordsAsPythonStrings = [str(x) for x in all_words]
 
         # get smallest type that can represent word offset
         self._wordOffsetType = self.compactNumpyTypeForValues([all_words.size * 2 + 10])
